@@ -36,8 +36,6 @@ static GstStaticPadTemplate gst_dvdlpcmdec_sink_template =
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("audio/x-private1-lpcm; "
-        "audio/x-private2-lpcm; "
-        "audio/x-private-ts-lpcm; "
         "audio/x-lpcm, "
         "width = (int) { 16, 20, 24 }, "
         "rate = (int) { 32000, 44100, 48000, 96000 }, "
@@ -58,42 +56,86 @@ GST_STATIC_PAD_TEMPLATE ("src",
         "channels = (int) [ 1, 8 ]")
     );
 
-#define gst_dvdlpcmdec_parent_class parent_class
-G_DEFINE_TYPE (GstDvdLpcmDec, gst_dvdlpcmdec, GST_TYPE_AUDIO_DECODER);
+/* DvdLpcmDec signals and args */
+enum
+{
+  /* FILL ME */
+  LAST_SIGNAL
+};
 
-static gboolean gst_dvdlpcmdec_set_format (GstAudioDecoder * bdec,
-    GstCaps * caps);
-static GstFlowReturn gst_dvdlpcmdec_parse (GstAudioDecoder * bdec,
-    GstAdapter * adapter, gint * offset, gint * len);
-static GstFlowReturn gst_dvdlpcmdec_handle_frame (GstAudioDecoder * bdec,
-    GstBuffer * buffer);
-static GstFlowReturn gst_dvdlpcmdec_chain (GstPad * pad, GstObject * parent,
-    GstBuffer * buffer);
+enum
+{
+  PROP_0
+      /* FILL ME */
+};
 
+static void gst_dvdlpcmdec_base_init (gpointer g_class);
+static void gst_dvdlpcmdec_class_init (GstDvdLpcmDecClass * klass);
+static void gst_dvdlpcmdec_init (GstDvdLpcmDec * dvdlpcmdec);
+
+static GstFlowReturn gst_dvdlpcmdec_chain_raw (GstPad * pad, GstObject * parent,
+    GstBuffer * buffer);
+static GstFlowReturn gst_dvdlpcmdec_chain_dvd (GstPad * pad, GstObject * parent,
+    GstBuffer * buffer);
+static gboolean gst_dvdlpcmdec_setcaps (GstPad * pad, GstCaps * caps);
+static gboolean dvdlpcmdec_sink_event (GstPad * pad, GstObject * parent,
+    GstEvent * event);
+
+static GstStateChangeReturn gst_dvdlpcmdec_change_state (GstElement * element,
+    GstStateChange transition);
+
+static GstElementClass *parent_class = NULL;
+
+GType
+gst_dvdlpcmdec_get_type (void)
+{
+  static GType dvdlpcmdec_type = 0;
+
+  if (!dvdlpcmdec_type) {
+    static const GTypeInfo dvdlpcmdec_info = {
+      sizeof (GstDvdLpcmDecClass),
+      gst_dvdlpcmdec_base_init,
+      NULL,
+      (GClassInitFunc) gst_dvdlpcmdec_class_init,
+      NULL,
+      NULL,
+      sizeof (GstDvdLpcmDec),
+      0,
+      (GInstanceInitFunc) gst_dvdlpcmdec_init,
+    };
+
+    dvdlpcmdec_type =
+        g_type_register_static (GST_TYPE_ELEMENT, "GstDvdLpcmDec",
+        &dvdlpcmdec_info, 0);
+  }
+  return dvdlpcmdec_type;
+}
 
 static void
-gst_dvdlpcmdec_class_init (GstDvdLpcmDecClass * klass)
+gst_dvdlpcmdec_base_init (gpointer g_class)
 {
-  GstElementClass *element_class;
-  GstAudioDecoderClass *gstbase_class;
+  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);
 
-  element_class = (GstElementClass *) klass;
-  gstbase_class = (GstAudioDecoderClass *) klass;
-
-  gstbase_class->set_format = GST_DEBUG_FUNCPTR (gst_dvdlpcmdec_set_format);
-  gstbase_class->parse = GST_DEBUG_FUNCPTR (gst_dvdlpcmdec_parse);
-  gstbase_class->handle_frame = GST_DEBUG_FUNCPTR (gst_dvdlpcmdec_handle_frame);
-
-  gst_element_class_add_static_pad_template (element_class,
-      &gst_dvdlpcmdec_sink_template);
-  gst_element_class_add_static_pad_template (element_class,
-      &gst_dvdlpcmdec_src_template);
+  gst_element_class_add_pad_template (element_class,
+      gst_static_pad_template_get (&gst_dvdlpcmdec_sink_template));
+  gst_element_class_add_pad_template (element_class,
+      gst_static_pad_template_get (&gst_dvdlpcmdec_src_template));
   gst_element_class_set_static_metadata (element_class,
       "DVD LPCM Audio decoder", "Codec/Decoder/Audio",
       "Decode DVD LPCM frames into standard PCM audio",
       "Jan Schmidt <jan@noraisin.net>, Michael Smith <msmith@fluendo.com>");
+}
 
-  GST_DEBUG_CATEGORY_INIT (dvdlpcm_debug, "dvdlpcmdec", 0, "DVD LPCM Decoder");
+static void
+gst_dvdlpcmdec_class_init (GstDvdLpcmDecClass * klass)
+{
+  GstElementClass *gstelement_class;
+
+  gstelement_class = (GstElementClass *) klass;
+
+  parent_class = g_type_class_peek_parent (klass);
+
+  gstelement_class->change_state = gst_dvdlpcmdec_change_state;
 }
 
 static void
@@ -103,28 +145,27 @@ gst_dvdlpcm_reset (GstDvdLpcmDec * dvdlpcmdec)
   dvdlpcmdec->dynamic_range = 0;
   dvdlpcmdec->emphasis = FALSE;
   dvdlpcmdec->mute = FALSE;
+  dvdlpcmdec->timestamp = GST_CLOCK_TIME_NONE;
 
   dvdlpcmdec->header = 0;
 
-  dvdlpcmdec->mode = GST_LPCM_UNKNOWN;
+  gst_segment_init (&dvdlpcmdec->segment, GST_FORMAT_UNDEFINED);
 }
 
 static void
 gst_dvdlpcmdec_init (GstDvdLpcmDec * dvdlpcmdec)
 {
+  dvdlpcmdec->sinkpad =
+      gst_pad_new_from_static_template (&gst_dvdlpcmdec_sink_template, "sink");
+  gst_pad_set_event_function (dvdlpcmdec->sinkpad, dvdlpcmdec_sink_event);
+  gst_element_add_pad (GST_ELEMENT (dvdlpcmdec), dvdlpcmdec->sinkpad);
+
+  dvdlpcmdec->srcpad =
+      gst_pad_new_from_static_template (&gst_dvdlpcmdec_src_template, "src");
+  gst_pad_use_fixed_caps (dvdlpcmdec->srcpad);
+  gst_element_add_pad (GST_ELEMENT (dvdlpcmdec), dvdlpcmdec->srcpad);
+
   gst_dvdlpcm_reset (dvdlpcmdec);
-
-  gst_audio_decoder_set_use_default_pad_acceptcaps (GST_AUDIO_DECODER_CAST
-      (dvdlpcmdec), TRUE);
-  GST_PAD_SET_ACCEPT_TEMPLATE (GST_AUDIO_DECODER_SINK_PAD (dvdlpcmdec));
-
-  /* retrieve and intercept base class chain.
-   * Quite HACKish, but that's dvd specs/caps for you,
-   * since one buffer needs to be split into 2 frames */
-  dvdlpcmdec->base_chain =
-      GST_PAD_CHAINFUNC (GST_AUDIO_DECODER_SINK_PAD (dvdlpcmdec));
-  gst_pad_set_chain_function (GST_AUDIO_DECODER_SINK_PAD (dvdlpcmdec),
-      GST_DEBUG_FUNCPTR (gst_dvdlpcmdec_chain));
 }
 
 static const GstAudioChannelPosition channel_positions[][8] = {
@@ -153,74 +194,6 @@ static const GstAudioChannelPosition channel_positions[][8] = {
   {GST_AUDIO_CHANNEL_POSITION_INVALID}
 };
 
-static const GstAudioChannelPosition bluray_channel_positions[][8] = {
-  /* invalid */
-  {GST_AUDIO_CHANNEL_POSITION_INVALID},
-  /* mono */
-  {GST_AUDIO_CHANNEL_POSITION_MONO},
-  /* invalid */
-  {GST_AUDIO_CHANNEL_POSITION_INVALID},
-  /* stereo */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-      GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT},
-  /* surround */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-      GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER},
-  /* 2.1 */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-      GST_AUDIO_CHANNEL_POSITION_REAR_CENTER},
-  /* 4.0 */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
-      GST_AUDIO_CHANNEL_POSITION_REAR_CENTER},
-  /* 2.2 */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_LEFT,
-      GST_AUDIO_CHANNEL_POSITION_SIDE_RIGHT},
-  /* 5.0 */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_LEFT,
-      GST_AUDIO_CHANNEL_POSITION_SIDE_RIGHT},
-  /* 5.1 */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_RIGHT,
-      GST_AUDIO_CHANNEL_POSITION_LFE1},
-  /* 7.0 */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_REAR_LEFT,
-      GST_AUDIO_CHANNEL_POSITION_REAR_RIGHT},
-  /* 7.1 */
-  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_FRONT_CENTER,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_SIDE_RIGHT,
-        GST_AUDIO_CHANNEL_POSITION_REAR_LEFT,
-        GST_AUDIO_CHANNEL_POSITION_REAR_RIGHT,
-      GST_AUDIO_CHANNEL_POSITION_LFE1},
-  /* invalid */
-  {GST_AUDIO_CHANNEL_POSITION_INVALID},
-  /* invalid */
-  {GST_AUDIO_CHANNEL_POSITION_INVALID},
-  /* invalid */
-  {GST_AUDIO_CHANNEL_POSITION_INVALID},
-  /* invalid */
-  {GST_AUDIO_CHANNEL_POSITION_INVALID}
-};
-
 static void
 gst_dvdlpcmdec_send_tags (GstDvdLpcmDec * dvdlpcmdec)
 {
@@ -236,90 +209,85 @@ gst_dvdlpcmdec_send_tags (GstDvdLpcmDec * dvdlpcmdec)
   taglist = gst_tag_list_new (GST_TAG_AUDIO_CODEC, "LPCM Audio",
       GST_TAG_BITRATE, bitrate, NULL);
 
-  gst_audio_decoder_merge_tags (GST_AUDIO_DECODER (dvdlpcmdec), taglist,
-      GST_TAG_MERGE_REPLACE);
-  gst_tag_list_unref (taglist);
+  gst_pad_push_event (dvdlpcmdec->srcpad, gst_event_new_tag (taglist));
 }
 
 static gboolean
-gst_dvdlpcmdec_set_output_format (GstDvdLpcmDec * dvdlpcmdec)
+gst_dvdlpcmdec_set_outcaps (GstDvdLpcmDec * dvdlpcmdec)
 {
   gboolean res = TRUE;
+  GstCaps *src_caps;
 
-  res = gst_audio_decoder_set_output_format (GST_AUDIO_DECODER (dvdlpcmdec),
-      &dvdlpcmdec->info);
+  /* Build caps to set on the src pad, which we know from the incoming caps */
+  src_caps = gst_audio_info_to_caps (&dvdlpcmdec->info);
+
+  res = gst_pad_set_caps (dvdlpcmdec->srcpad, src_caps);
   if (res) {
-    GST_DEBUG_OBJECT (dvdlpcmdec, "Successfully set output format");
+    GST_DEBUG_OBJECT (dvdlpcmdec, "Successfully set output caps: %"
+        GST_PTR_FORMAT, src_caps);
 
     gst_dvdlpcmdec_send_tags (dvdlpcmdec);
   } else {
-    GST_DEBUG_OBJECT (dvdlpcmdec, "Failed to set output format");
+    GST_DEBUG_OBJECT (dvdlpcmdec, "Failed to set output caps: %"
+        GST_PTR_FORMAT, src_caps);
   }
 
+  gst_caps_unref (src_caps);
+
   return res;
 }
 
 static void
 gst_dvdlpcmdec_update_audio_formats (GstDvdLpcmDec * dec, gint channels,
-    gint rate, GstAudioFormat format, guint8 channel_indicator,
-    const GstAudioChannelPosition positions[][8])
+    gint rate, GstAudioFormat format)
 {
-  GST_DEBUG_OBJECT (dec, "got channels = %d, rate = %d, format = %d", channels,
+  GST_DEBUG_OBJECT (dec, "got channles = %d, rate = %d, format = %d", channels,
       rate, format);
 
   /* Reorder the channel positions and set the default into for the audio */
   if (channels < 9
-      && positions[channel_indicator][0] !=
+      && channel_positions[channels - 1][0] !=
       GST_AUDIO_CHANNEL_POSITION_INVALID) {
     const GstAudioChannelPosition *position;
     GstAudioChannelPosition sorted_position[8];
     guint c;
 
-    position = positions[channel_indicator];
+    position = channel_positions[channels - 1];
+    dec->lpcm_layout = position;
     for (c = 0; c < channels; ++c)
       sorted_position[c] = position[c];
     gst_audio_channel_positions_to_valid_order (sorted_position, channels);
     gst_audio_info_set_format (&dec->info, format, rate, channels,
         sorted_position);
-    if (memcmp (position, sorted_position,
-            channels * sizeof (position[0])) != 0)
-      dec->lpcm_layout = position;
-    else
-      dec->lpcm_layout = NULL;
   } else {
     gst_audio_info_set_format (&dec->info, format, rate, channels, NULL);
   }
 }
 
 static gboolean
-gst_dvdlpcmdec_set_format (GstAudioDecoder * bdec, GstCaps * caps)
+gst_dvdlpcmdec_setcaps (GstPad * pad, GstCaps * caps)
 {
-  GstDvdLpcmDec *dvdlpcmdec = GST_DVDLPCMDEC (bdec);
   GstStructure *structure;
   gboolean res = TRUE;
+  GstDvdLpcmDec *dvdlpcmdec;
   GstAudioFormat format;
   gint rate, channels, width;
 
-  gst_dvdlpcm_reset (dvdlpcmdec);
+  g_return_val_if_fail (caps != NULL, FALSE);
+  g_return_val_if_fail (pad != NULL, FALSE);
+
+  dvdlpcmdec = GST_DVDLPCMDEC (gst_pad_get_parent (pad));
 
   structure = gst_caps_get_structure (caps, 0);
 
   /* If we have the DVD structured LPCM (including header) then we wait
    * for incoming data before creating the output pad caps */
   if (gst_structure_has_name (structure, "audio/x-private1-lpcm")) {
-    dvdlpcmdec->mode = GST_LPCM_DVD;
-    goto done;
-  }
-  if (gst_structure_has_name (structure, "audio/x-private2-lpcm")) {
-    dvdlpcmdec->mode = GST_LPCM_1394;
-    goto done;
-  }
-  if (gst_structure_has_name (structure, "audio/x-private-ts-lpcm")) {
-    dvdlpcmdec->mode = GST_LPCM_BLURAY;
+    gst_pad_set_chain_function (dvdlpcmdec->sinkpad, gst_dvdlpcmdec_chain_dvd);
     goto done;
   }
 
-  dvdlpcmdec->mode = GST_LPCM_RAW;
+  gst_pad_set_chain_function (dvdlpcmdec->sinkpad, gst_dvdlpcmdec_chain_raw);
 
   res &= gst_structure_get_int (structure, "rate", &rate);
   res &= gst_structure_get_int (structure, "channels", &channels);
@@ -346,24 +314,63 @@ gst_dvdlpcmdec_set_format (GstAudioDecoder * bdec, GstCaps * caps)
       break;
   }
 
-  gst_dvdlpcmdec_update_audio_formats (dvdlpcmdec, channels, rate, format,
-      channels - 1, channel_positions);
+  gst_dvdlpcmdec_update_audio_formats (dvdlpcmdec, channels, rate, format);
 
   dvdlpcmdec->width = width;
 
-  res = gst_dvdlpcmdec_set_output_format (dvdlpcmdec);
+  res = gst_dvdlpcmdec_set_outcaps (dvdlpcmdec);
 
 done:
+  gst_object_unref (dvdlpcmdec);
   return res;
 
   /* ERRORS */
 caps_parse_error:
   {
     GST_DEBUG_OBJECT (dvdlpcmdec, "Couldn't get parameters; missing caps?");
+    gst_object_unref (dvdlpcmdec);
     return FALSE;
   }
 }
 
+static void
+update_timestamps (GstDvdLpcmDec * dvdlpcmdec, GstBuffer * buf, int samples)
+{
+  gboolean take_buf_ts = FALSE;
+  gint rate;
+
+  rate = GST_AUDIO_INFO_RATE (&dvdlpcmdec->info);
+
+  GST_BUFFER_DURATION (buf) = gst_util_uint64_scale (samples, GST_SECOND, rate);
+
+  if (GST_BUFFER_TIMESTAMP_IS_VALID (buf)) {
+    if (GST_CLOCK_TIME_IS_VALID (dvdlpcmdec->timestamp)) {
+      GstClockTimeDiff one_sample = GST_SECOND / rate;
+      GstClockTimeDiff diff = GST_CLOCK_DIFF (GST_BUFFER_TIMESTAMP (buf),
+          dvdlpcmdec->timestamp);
+
+      if (diff > one_sample || diff < -one_sample)
+        take_buf_ts = TRUE;
+    } else {
+      take_buf_ts = TRUE;
+    }
+  } else if (!GST_CLOCK_TIME_IS_VALID (dvdlpcmdec->timestamp)) {
+    dvdlpcmdec->timestamp = 0;
+  }
+
+  if (take_buf_ts) {
+    /* Take buffer timestamp */
+    dvdlpcmdec->timestamp = GST_BUFFER_TIMESTAMP (buf);
+  } else {
+    GST_BUFFER_TIMESTAMP (buf) = dvdlpcmdec->timestamp;
+  }
+
+  dvdlpcmdec->timestamp += GST_BUFFER_DURATION (buf);
+
+  GST_LOG_OBJECT (dvdlpcmdec, "Updated timestamp to %" GST_TIME_FORMAT,
+      GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (buf)));
+}
+
 static void
 parse_header (GstDvdLpcmDec * dec, guint32 header)
 {
@@ -419,45 +426,77 @@ parse_header (GstDvdLpcmDec * dec, guint32 header)
   /* And, of course, the number of channels (up to 8) */
   channels = ((header >> 8) & 0x7) + 1;
 
-  gst_dvdlpcmdec_update_audio_formats (dec, channels, rate, format,
-      channels - 1, channel_positions);
+  gst_dvdlpcmdec_update_audio_formats (dec, channels, rate, format);
 }
 
 static GstFlowReturn
-gst_dvdlpcmdec_chain (GstPad * pad, GstObject * parent, GstBuffer * buf)
+gst_dvdlpcmdec_chain_dvd (GstPad * pad, GstObject * parent, GstBuffer * buf)
 {
-  GstDvdLpcmDec *dvdlpcmdec = GST_DVDLPCMDEC (parent);
-  guint8 data[2];
+  GstDvdLpcmDec *dvdlpcmdec;
+  GstMapInfo map;
+  guint8 *data;
   gsize size;
   guint first_access;
+  guint32 header;
   GstBuffer *subbuf;
   GstFlowReturn ret = GST_FLOW_OK;
   gint off, len;
+  gint rate, channels;
 
-  if (dvdlpcmdec->mode != GST_LPCM_DVD)
-    return dvdlpcmdec->base_chain (pad, parent, buf);
+  dvdlpcmdec = GST_DVDLPCMDEC (parent);
+
+  gst_buffer_map (buf, &map, GST_MAP_READ);
+  data = map.data;
+  size = map.size;
 
-  size = gst_buffer_get_size (buf);
   if (size < 5)
     goto too_small;
 
-  gst_buffer_extract (buf, 0, data, 2);
+  /* We have a 5 byte header, now.
+   * The first two bytes are a (big endian) 16 bit offset into our buffer.
+   * The buffer timestamp refers to this offset.
+   *
+   * The other three bytes are a (big endian) number in which the header is
+   * encoded.
+   */
   first_access = (data[0] << 8) | data[1];
   if (first_access > size)
     goto invalid_data;
 
-  /* After first_access, we have an additional 3 bytes of header data; this
-   * is included within the value of first_access.
+  /* Don't keep the 'frame number' low 5 bits of the first byte */
+  header = ((data[2] & 0xC0) << 16) | (data[3] << 8) | data[4];
+
+  /* see if we have a new header */
+  if (header != dvdlpcmdec->header) {
+    parse_header (dvdlpcmdec, header);
+
+    if (!gst_dvdlpcmdec_set_outcaps (dvdlpcmdec))
+      goto negotiation_failed;
+
+    dvdlpcmdec->header = header;
+  }
+
+  GST_LOG_OBJECT (dvdlpcmdec, "first_access %d, buffer length %" G_GSIZE_FORMAT,
+      first_access, size);
+
+  rate = GST_AUDIO_INFO_RATE (&dvdlpcmdec->info);
+  channels = GST_AUDIO_INFO_CHANNELS (&dvdlpcmdec->info);
+
+  /* After first_access, we have an additional 3 bytes of data we've parsed and
+   * don't want to handle; this is included within the value of first_access.
    * So a first_access value of between 1 and 3 is just broken, we treat that
    * the same as zero. first_access == 4 means we only need to create a single
    * sub-buffer, greater than that we need to create two. */
 
-  /* skip access unit bytes */
-  off = 2;
+  /* skip access unit bytes and info */
+  off = 5;
 
   if (first_access > 4) {
+    guint samples = 0;
+    GstClockTime ts;
+
     /* length of first buffer */
-    len = first_access - 1;
+    len = first_access - 4;
 
     GST_LOG_OBJECT (dvdlpcmdec, "Creating first sub-buffer off %d, len %d",
         off, len);
@@ -467,8 +506,36 @@ gst_dvdlpcmdec_chain (GstPad * pad, GstObject * parent, GstBuffer * buf)
       goto bad_first_access;
 
     subbuf = gst_buffer_copy_region (buf, GST_BUFFER_COPY_ALL, off, len);
-    GST_BUFFER_TIMESTAMP (subbuf) = GST_CLOCK_TIME_NONE;
-    ret = dvdlpcmdec->base_chain (pad, parent, subbuf);
+
+    /* If we don't have a stored timestamp from the last packet,
+     * (it's straight after a new-segment, but we have one on the
+     * first access buffer, then calculate the timestamp to align
+     * this buffer to just before the first_access buffer */
+    if (!GST_CLOCK_TIME_IS_VALID (dvdlpcmdec->timestamp) &&
+        GST_BUFFER_TIMESTAMP_IS_VALID (buf)) {
+      switch (dvdlpcmdec->width) {
+        case 16:
+          samples = len / channels / 2;
+          break;
+        case 20:
+          samples = (len / channels) * 2 / 5;
+          break;
+        case 24:
+          samples = len / channels / 3;
+          break;
+      }
+    }
+    if (samples != 0) {
+      ts = gst_util_uint64_scale (samples, GST_SECOND, rate);
+      if (ts < GST_BUFFER_TIMESTAMP (buf))
+        GST_BUFFER_TIMESTAMP (subbuf) = GST_BUFFER_TIMESTAMP (buf) - ts;
+      else
+        GST_BUFFER_TIMESTAMP (subbuf) = 0;
+    } else {
+      GST_BUFFER_TIMESTAMP (subbuf) = GST_CLOCK_TIME_NONE;
+    }
+
+    ret = gst_dvdlpcmdec_chain_raw (pad, parent, subbuf);
     if (ret != GST_FLOW_OK)
       goto done;
 
@@ -480,14 +547,10 @@ gst_dvdlpcmdec_chain (GstPad * pad, GstObject * parent, GstBuffer * buf)
         len);
 
     if (len > 0) {
-      GstMemory *header, *tmp;
       subbuf = gst_buffer_copy_region (buf, GST_BUFFER_COPY_ALL, off, len);
-      tmp = gst_buffer_peek_memory (buf, 0);
-      header = gst_memory_copy (tmp, 2, 3);
-      gst_buffer_prepend_memory (subbuf, header);
       GST_BUFFER_TIMESTAMP (subbuf) = GST_BUFFER_TIMESTAMP (buf);
 
-      ret = dvdlpcmdec->base_chain (pad, parent, subbuf);
+      ret = gst_dvdlpcmdec_chain_raw (pad, parent, subbuf);
     }
   } else {
     GST_LOG_OBJECT (dvdlpcmdec,
@@ -495,10 +558,11 @@ gst_dvdlpcmdec_chain (GstPad * pad, GstObject * parent, GstBuffer * buf)
         size - off);
     subbuf = gst_buffer_copy_region (buf, GST_BUFFER_COPY_ALL, off, size - off);
     GST_BUFFER_TIMESTAMP (subbuf) = GST_BUFFER_TIMESTAMP (buf);
-    ret = dvdlpcmdec->base_chain (pad, parent, subbuf);
+    ret = gst_dvdlpcmdec_chain_raw (pad, parent, subbuf);
   }
 
 done:
+  gst_buffer_unmap (buf, &map);
   gst_buffer_unref (buf);
 
   return ret;
@@ -521,6 +585,13 @@ invalid_data:
     ret = GST_FLOW_OK;
     goto done;
   }
+negotiation_failed:
+  {
+    GST_ELEMENT_ERROR (dvdlpcmdec, STREAM, FORMAT, (NULL),
+        ("Failed to configure output format"));
+    ret = GST_FLOW_NOT_NEGOTIATED;
+    goto done;
+  }
 bad_first_access:
   {
     GST_WARNING_OBJECT (pad, "Bad first_access parameter in buffer");
@@ -533,315 +604,15 @@ bad_first_access:
 }
 
 static GstFlowReturn
-gst_dvdlpcmdec_parse_dvd (GstDvdLpcmDec * dvdlpcmdec, GstAdapter * adapter,
-    gint * offset, gint * len)
+gst_dvdlpcmdec_chain_raw (GstPad * pad, GstObject * parent, GstBuffer * buf)
 {
-  guint32 header;
-  const guint8 *data;
-
-  data = (const guint8 *) gst_adapter_map (adapter, 3);
-  if (!data)
-    goto too_small;
-
-  /* Don't keep the 'frame number' low 5 bits of the first byte */
-  header = ((data[0] & 0xC0) << 16) | (data[1] << 8) | data[2];
-
-  gst_adapter_unmap (adapter);
-
-  /* see if we have a new header */
-  if (header != dvdlpcmdec->header) {
-    parse_header (dvdlpcmdec, header);
-
-    if (!gst_dvdlpcmdec_set_output_format (dvdlpcmdec))
-      goto negotiation_failed;
-
-    dvdlpcmdec->header = header;
-  }
-
-  *offset = 3;
-  *len = gst_adapter_available (adapter) - 3;
-
-  return GST_FLOW_OK;
-
-  /* ERRORS */
-too_small:
-  {
-    /* Buffer is too small */
-    GST_ELEMENT_WARNING (dvdlpcmdec, STREAM, DECODE,
-        ("Invalid data found parsing LPCM packet"),
-        ("LPCM packet was too small. Dropping"));
-    *offset = gst_adapter_available (adapter);
-    return GST_FLOW_EOS;
-  }
-negotiation_failed:
-  {
-    GST_ELEMENT_ERROR (dvdlpcmdec, STREAM, FORMAT, (NULL),
-        ("Failed to configure output format"));
-    return GST_FLOW_NOT_NEGOTIATED;
-  }
-}
-
-static GstFlowReturn
-gst_dvdlpcmdec_parse_bluray (GstDvdLpcmDec * dvdlpcmdec, GstAdapter * adapter,
-    gint * offset, gint * len)
-{
-  guint32 header;
-  const guint8 *data;
-  guint8 channel_indicator;
-
-  data = (const guint8 *) gst_adapter_map (adapter, 4);
-  if (!data)
-    goto too_small;
-
-  header = GST_READ_UINT32_BE (data);
-
-  gst_adapter_unmap (adapter);
-
-  /* see if we have a new header */
-  if (header != dvdlpcmdec->header) {
-    GstAudioFormat format;
-    gint rate, channels;
-
-    switch ((header >> 6) & 0x3) {
-      case 0x1:
-        format = GST_AUDIO_FORMAT_S16BE;
-        dvdlpcmdec->width = 16;
-        break;
-      case 0x2:
-        format = GST_AUDIO_FORMAT_S24BE;
-        dvdlpcmdec->width = 20;
-        break;
-      case 0x3:
-        format = GST_AUDIO_FORMAT_S24BE;
-        dvdlpcmdec->width = 24;
-        break;
-      default:
-        format = GST_AUDIO_FORMAT_UNKNOWN;
-        dvdlpcmdec->width = 0;
-        GST_WARNING ("Invalid sample depth!");
-        break;
-    }
-
-    switch ((header >> 8) & 0xf) {
-      case 0x1:
-        rate = 48000;
-        break;
-      case 0x4:
-        rate = 96000;
-        break;
-      case 0x5:
-        rate = 192000;
-        break;
-      default:
-        rate = 0;
-        GST_WARNING ("Invalid audio sampling frequency!");
-        break;
-    }
-    channel_indicator = (header >> 12) & 0xf;
-    switch (channel_indicator) {
-      case 0x1:
-        channels = 1;
-        break;
-      case 0x3:
-        channels = 2;
-        break;
-      case 0x4:
-      case 0x5:
-        channels = 3;
-        break;
-      case 0x6:
-      case 0x7:
-        channels = 4;
-        break;
-      case 0x8:
-        channels = 5;
-        break;
-      case 0x9:
-        channels = 6;
-        break;
-      case 0xa:
-        channels = 7;
-        break;
-      case 0xb:
-        channels = 8;
-        break;
-      default:
-        channels = 0;
-        GST_WARNING ("Invalid number of audio channels!");
-        goto negotiation_failed;
-    }
-    GST_DEBUG_OBJECT (dvdlpcmdec, "got channels %d rate %d format %s",
-        channels, rate, gst_audio_format_to_string (format));
-
-    gst_dvdlpcmdec_update_audio_formats (dvdlpcmdec, channels, rate, format,
-        channel_indicator, bluray_channel_positions);
-
-    if (!gst_dvdlpcmdec_set_output_format (dvdlpcmdec))
-      goto negotiation_failed;
-
-    dvdlpcmdec->header = header;
-  }
-
-  *offset = 4;
-  *len = gst_adapter_available (adapter) - 4;
-
-  return GST_FLOW_OK;
-
-  /* ERRORS */
-too_small:
-  {
-    /* Buffer is too small */
-    GST_ELEMENT_WARNING (dvdlpcmdec, STREAM, DECODE,
-        ("Invalid data found parsing LPCM packet"),
-        ("LPCM packet was too small. Dropping"));
-    *offset = gst_adapter_available (adapter);
-    return GST_FLOW_EOS;
-  }
-negotiation_failed:
-  {
-    GST_ELEMENT_ERROR (dvdlpcmdec, STREAM, FORMAT, (NULL),
-        ("Failed to configure output format"));
-    return GST_FLOW_NOT_NEGOTIATED;
-  }
-
-}
-
-static GstFlowReturn
-gst_dvdlpcmdec_parse_1394 (GstDvdLpcmDec * dvdlpcmdec, GstAdapter * adapter,
-    gint * offset, gint * len)
-{
-  guint32 header;
-  const guint8 *data;
-
-  data = (const guint8 *) gst_adapter_map (adapter, 4);
-  if (!data)
-    goto too_small;
-
-  header = GST_READ_UINT32_BE (data);
-
-  gst_adapter_unmap (adapter);
-
-  /* see if we have a new header */
-  if (header != dvdlpcmdec->header) {
-    GstAudioFormat format;
-    gint rate, channels;
-
-    if (header >> 24 != 0xa0)
-      goto invalid_data;
-
-    switch ((header >> 6) & 0x3) {
-      case 0x0:
-        format = GST_AUDIO_FORMAT_S16BE;
-        dvdlpcmdec->width = 16;
-        break;
-      default:
-        format = GST_AUDIO_FORMAT_UNKNOWN;
-        dvdlpcmdec->width = 0;
-        GST_WARNING ("Invalid quantization word length!");
-        break;
-    }
-
-    switch ((header >> 3) & 0x7) {
-      case 0x1:
-        rate = 44100;
-        break;
-      case 0x2:
-        rate = 48000;
-        break;
-      default:
-        rate = 0;
-        GST_WARNING ("Invalid audio sampling frequency!");
-        break;
-    }
-    switch (header & 0x7) {
-      case 0x0:                /* 2 channels dual-mono */
-      case 0x1:                /* 2 channels stereo */
-        channels = 2;
-        break;
-      default:
-        channels = 0;
-        GST_WARNING ("Invalid number of audio channels!");
-        goto negotiation_failed;
-    }
-
-    gst_dvdlpcmdec_update_audio_formats (dvdlpcmdec, channels, rate, format,
-        channels - 1, channel_positions);
-
-    if (!gst_dvdlpcmdec_set_output_format (dvdlpcmdec))
-      goto negotiation_failed;
-
-    dvdlpcmdec->header = header;
-  }
-
-  *offset = 4;
-  *len = gst_adapter_available (adapter) - 4;
-
-  return GST_FLOW_OK;
-
-  /* ERRORS */
-too_small:
-  {
-    /* Buffer is too small */
-    GST_ELEMENT_WARNING (dvdlpcmdec, STREAM, DECODE,
-        ("Invalid data found parsing LPCM packet"),
-        ("LPCM packet was too small. Dropping"));
-    *offset = gst_adapter_available (adapter);
-    return GST_FLOW_EOS;
-  }
-invalid_data:
-  {
-    GST_ELEMENT_WARNING (dvdlpcmdec, STREAM, DECODE,
-        ("Invalid data found parsing LPCM packet"),
-        ("LPCM packet contains invalid sub_stream_id."));
-    return GST_FLOW_ERROR;
-  }
-negotiation_failed:
-  {
-    GST_ELEMENT_ERROR (dvdlpcmdec, STREAM, FORMAT, (NULL),
-        ("Failed to configure output format"));
-    return GST_FLOW_NOT_NEGOTIATED;
-  }
-}
-
-static GstFlowReturn
-gst_dvdlpcmdec_parse (GstAudioDecoder * bdec, GstAdapter * adapter,
-    gint * offset, gint * len)
-{
-  GstDvdLpcmDec *dvdlpcmdec = GST_DVDLPCMDEC (bdec);
-
-  switch (dvdlpcmdec->mode) {
-    case GST_LPCM_UNKNOWN:
-      return GST_FLOW_NOT_NEGOTIATED;
-
-    case GST_LPCM_RAW:
-      *offset = 0;
-      *len = gst_adapter_available (adapter);
-      return GST_FLOW_OK;
-
-    case GST_LPCM_DVD:
-      return gst_dvdlpcmdec_parse_dvd (dvdlpcmdec, adapter, offset, len);
-
-    case GST_LPCM_1394:
-      return gst_dvdlpcmdec_parse_1394 (dvdlpcmdec, adapter, offset, len);
-
-    case GST_LPCM_BLURAY:
-      return gst_dvdlpcmdec_parse_bluray (dvdlpcmdec, adapter, offset, len);
-  }
-  return GST_FLOW_ERROR;
-}
-
-static GstFlowReturn
-gst_dvdlpcmdec_handle_frame (GstAudioDecoder * bdec, GstBuffer * buf)
-{
-  GstDvdLpcmDec *dvdlpcmdec = GST_DVDLPCMDEC (bdec);
+  GstDvdLpcmDec *dvdlpcmdec;
   gsize size;
   GstFlowReturn ret;
   guint samples = 0;
   gint rate, channels;
 
-  /* no fancy draining */
-  if (G_UNLIKELY (!buf))
-    return GST_FLOW_OK;
+  dvdlpcmdec = GST_DVDLPCMDEC (parent);
 
   size = gst_buffer_get_size (buf);
 
@@ -855,6 +626,9 @@ gst_dvdlpcmdec_handle_frame (GstAudioDecoder * bdec, GstBuffer * buf)
   if (rate == 0)
     goto not_negotiated;
 
+  if (GST_BUFFER_TIMESTAMP_IS_VALID (buf))
+    dvdlpcmdec->timestamp = GST_BUFFER_TIMESTAMP (buf);
+
   /* We don't currently do anything at all regarding emphasis, mute or
    * dynamic_range - I'm not sure what they're for */
   switch (dvdlpcmdec->width) {
@@ -865,8 +639,7 @@ gst_dvdlpcmdec_handle_frame (GstAudioDecoder * bdec, GstBuffer * buf)
       samples = size / channels / 2;
       if (samples < 1)
         goto drop;
-
-      gst_buffer_ref (buf);
+      buf = gst_buffer_make_writable (buf);
       break;
     }
     case 20:
@@ -915,6 +688,7 @@ gst_dvdlpcmdec_handle_frame (GstAudioDecoder * bdec, GstBuffer * buf)
       }
       gst_buffer_unmap (outbuf, &destmap);
       gst_buffer_unmap (buf, &srcmap);
+      gst_buffer_unref (buf);
       buf = outbuf;
       break;
     }
@@ -924,57 +698,51 @@ gst_dvdlpcmdec_handle_frame (GstAudioDecoder * bdec, GstBuffer * buf)
        * and last byte are already correct */
       guint count = size / 12;
       gint i;
-      GstMapInfo srcmap, destmap;
-      guint8 *src, *dest;
-      GstBuffer *outbuf;
+      GstMapInfo map;
+      guint8 *ptr;
 
       samples = size / channels / 3;
 
       if (samples < 1)
         goto drop;
 
-      outbuf = gst_buffer_new_allocate (NULL, size, NULL);
-      gst_buffer_copy_into (outbuf, buf, GST_BUFFER_COPY_TIMESTAMPS, 0, -1);
+      /* Ensure our output buffer is writable */
+      buf = gst_buffer_make_writable (buf);
 
-      gst_buffer_map (buf, &srcmap, GST_MAP_READ);
-      gst_buffer_map (outbuf, &destmap, GST_MAP_READWRITE);
-      src = srcmap.data;
-      dest = destmap.data;
+      gst_buffer_map (buf, &map, GST_MAP_READWRITE);
+      ptr = map.data;
 
       for (i = 0; i < count; i++) {
-        dest[0] = src[0];
-        dest[1] = src[1];
-        dest[11] = src[11];
-        dest[10] = src[7];
-        dest[7] = src[5];
-        dest[5] = src[9];
-        dest[9] = src[6];
-        dest[6] = src[4];
-        dest[4] = src[3];
-        dest[3] = src[2];
-        dest[2] = src[8];
-        dest[8] = src[10];
+        guint8 tmp;
 
-        src += 12;
-        dest += 12;
+        tmp = ptr[10];
+        ptr[10] = ptr[7];
+        ptr[7] = ptr[5];
+        ptr[5] = ptr[9];
+        ptr[9] = ptr[6];
+        ptr[6] = ptr[4];
+        ptr[4] = ptr[3];
+        ptr[3] = ptr[2];
+        ptr[2] = ptr[8];
+        ptr[8] = tmp;
+
+        ptr += 12;
       }
-      gst_buffer_unmap (outbuf, &destmap);
-      gst_buffer_unmap (buf, &srcmap);
-      buf = outbuf;
+      gst_buffer_unmap (buf, &map);
       break;
     }
     default:
       goto invalid_width;
   }
 
-  if (dvdlpcmdec->lpcm_layout) {
-    buf = gst_buffer_make_writable (buf);
+  update_timestamps (dvdlpcmdec, buf, samples);
+
+  if (dvdlpcmdec->lpcm_layout)
     gst_audio_buffer_reorder_channels (buf, dvdlpcmdec->info.finfo->format,
         dvdlpcmdec->info.channels, dvdlpcmdec->lpcm_layout,
         dvdlpcmdec->info.position);
-  }
 
-  ret = gst_audio_decoder_finish_frame (bdec, buf, 1);
+  ret = gst_pad_push (dvdlpcmdec->srcpad, buf);
 
 done:
   return ret;
@@ -984,6 +752,7 @@ drop:
   {
     GST_DEBUG_OBJECT (dvdlpcmdec,
         "Buffer of size %" G_GSIZE_FORMAT " is too small. Dropping", size);
+    gst_buffer_unref (buf);
     ret = GST_FLOW_OK;
     goto done;
   }
@@ -991,6 +760,7 @@ not_negotiated:
   {
     GST_ELEMENT_ERROR (dvdlpcmdec, STREAM, FORMAT, (NULL),
         ("Buffer pushed before negotiation"));
+    gst_buffer_unref (buf);
     ret = GST_FLOW_NOT_NEGOTIATED;
     goto done;
   }
@@ -998,14 +768,88 @@ invalid_width:
   {
     GST_ELEMENT_ERROR (dvdlpcmdec, STREAM, WRONG_TYPE, (NULL),
         ("Invalid sample width configured"));
+    gst_buffer_unref (buf);
     ret = GST_FLOW_NOT_NEGOTIATED;
     goto done;
   }
 }
 
+static gboolean
+dvdlpcmdec_sink_event (GstPad * pad, GstObject * parent, GstEvent * event)
+{
+  GstDvdLpcmDec *dvdlpcmdec;
+  gboolean res;
+
+  dvdlpcmdec = GST_DVDLPCMDEC (parent);
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_CAPS:
+    {
+      GstCaps *caps;
+
+      gst_event_parse_caps (event, &caps);
+      res = gst_dvdlpcmdec_setcaps (pad, caps);
+      gst_event_unref (event);
+      break;
+    }
+    case GST_EVENT_SEGMENT:
+    {
+      GstSegment seg;
+
+      gst_event_copy_segment (event, &seg);
+
+      GST_DEBUG_OBJECT (dvdlpcmdec, "segment %" GST_SEGMENT_FORMAT, &seg);
+
+      dvdlpcmdec->segment = seg;
+
+      if (seg.format == GST_FORMAT_TIME) {
+        dvdlpcmdec->timestamp = GST_CLOCK_TIME_NONE;
+      } else {
+        dvdlpcmdec->timestamp = 0;
+      }
+      res = gst_pad_push_event (dvdlpcmdec->srcpad, event);
+      break;
+    }
+    case GST_EVENT_FLUSH_STOP:
+      gst_segment_init (&dvdlpcmdec->segment, GST_FORMAT_UNDEFINED);
+      res = gst_pad_push_event (dvdlpcmdec->srcpad, event);
+      break;
+    default:
+      res = gst_pad_push_event (dvdlpcmdec->srcpad, event);
+      break;
+  }
+
+  return res;
+}
+
+static GstStateChangeReturn
+gst_dvdlpcmdec_change_state (GstElement * element, GstStateChange transition)
+{
+  GstDvdLpcmDec *dvdlpcmdec = GST_DVDLPCMDEC (element);
+  GstStateChangeReturn res;
+
+  switch (transition) {
+    case GST_STATE_CHANGE_READY_TO_PAUSED:
+      gst_dvdlpcm_reset (dvdlpcmdec);
+      break;
+    default:
+      break;
+  }
+
+  res = parent_class->change_state (element, transition);
+
+  switch (transition) {
+    default:
+      break;
+  }
+
+  return res;
+}
+
 static gboolean
 plugin_init (GstPlugin * plugin)
 {
+  GST_DEBUG_CATEGORY_INIT (dvdlpcm_debug, "dvdlpcmdec", 0, "DVD LPCM Decoder");
 
   if (!gst_element_register (plugin, "dvdlpcmdec", GST_RANK_PRIMARY,
           GST_TYPE_DVDLPCMDEC)) {
